# -*- coding: utf-8 -*-
# Time       : 2023/10/14 19:14
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import os
import shutil
from pathlib import Path

import cv2
import onnxruntime
from tqdm import tqdm

from hcaptcha_challenger.onnx.modelhub import ModelHub
from hcaptcha_challenger.onnx.modelhub import request_resource
from hcaptcha_challenger.onnx.yolo import YOLOv8


class CcYOLO:
    model_url = "https://github.com/QIN2DIM/hcaptcha-challenger/releases/download/model/"

    def __init__(self, model_name: str, images_absolute_dir: Path, this_dir: Path):
        assert images_absolute_dir

        input_dir = images_absolute_dir
        if not input_dir.absolute():
            raise ValueError("`images_absolute_dir` Should be an absolute path")
        if not input_dir.is_dir():
            raise FileNotFoundError("`images_absolute_dir` pointed to does not exist")

        self.modelhub = ModelHub.from_github_repo()
        self.modelhub.parse_objects()

        self.model_name = model_name
        self.input_dir = input_dir

        self.output_dir = this_dir
        if input_dir.parent.name:
            self.output_dir = self.output_dir.joinpath(input_dir.parent.name)
        self.output_dir = self.output_dir.joinpath(input_dir.name)
        self.output_miss_dir = self.output_dir.joinpath("miss")

        self.output_miss_dir.mkdir(parents=True, exist_ok=True)

    @property
    def desc_out(self):
        return f'"{self.output_dir.parent.name}/{self.output_dir.name}"'

    @property
    def desc_in(self):
        return f'"{self.input_dir.parent.name}/{self.input_dir.name}"'

    def get_model(self) -> YOLOv8 | None:
        classes = self.modelhub.ashes_of_war.get(self.model_name)
        if not classes:
            raise AttributeError(f"Model name not found - {self.model_name=}")

        print(f">> Match model - {self.model_name=}")
        model_path = self.modelhub.models_dir.joinpath(self.model_name)
        if not model_path.exists():
            request_resource(self.model_url + self.model_name, model_path)

        session = onnxruntime.InferenceSession(
            model_path, providers=onnxruntime.get_available_providers()
        )
        detector = YOLOv8.from_pluggable_model(session, classes)

        return detector

    def spawn(self):
        """
        ```python
        import os
        import sys
        from pathlib import Path

        from hcaptcha_challenger.components.yolo_mocker import CcYOLO


        def run():
            model_name = "head_of_the_animal_2310_yolov8s.onnx"
            images_dir = "image_label_area_select/please click on the head of the animal/default"

            this_dir = Path(__file__).parent
            output_dir = this_dir.joinpath("yolo_mocker")
            images_dir = this_dir.joinpath(images_dir).absolute()

            ccy = CcYOLO(model_name, images_dir, output_dir)
            ccy.spawn()

            if "win32" in sys.platform:
                os.startfile(ccy.output_dir)


        if __name__ == "__main__":
            run()
        ```
        :return:
        """

        def draw():
            alts = sorted(results, key=lambda x: x[-1])
            text, ps, pe, _ = alts[-1]
            image = cv2.imread(str(image_path))
            pt1 = int(ps[0]), int(ps[1])
            pt2 = int(pe[0]), int(pe[1])
            cv2.rectangle(image, pt1, pt2, (87, 241, 126), 2)
            cv2.imwrite(str(output_path), image)

        if not (detector := self.get_model()):
            return

        pending_image_paths = []
        for image_name in os.listdir(self.input_dir):
            image_path = self.input_dir.joinpath(image_name)
            if image_path.is_file() and not self.output_dir.joinpath(image_name).exists():
                pending_image_paths.append(image_path)

        total = len(pending_image_paths)
        handle, miss = 0, 0

        with tqdm(total=total, desc=f"Labeling | {self.desc_in}") as progress:
            for image_path in pending_image_paths:
                results = detector(image_path, shape_type="bounding_box")
                progress.update(1)
                if not results:
                    output_miss_path = self.output_miss_dir.joinpath(image_path.name)
                    shutil.copyfile(image_path, output_miss_path)
                    miss += 1
                    continue
                output_path = self.output_dir.joinpath(image_path.name)
                draw()
                handle += 1
        print(f">> Statistic - {total=} {handle=} {miss=} - input --> {self.desc_in}")



# -*- coding: utf-8 -*-
# Time       : 2023/8/19 17:53
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from __future__ import annotations

import re
from contextlib import suppress
from pathlib import Path
from typing import List

import cv2
from loguru import logger

from hcaptcha_challenger.components.prompt_handler import label_cleaning, split_prompt_message
from hcaptcha_challenger.onnx.modelhub import ModelHub
from hcaptcha_challenger.onnx.resnet import ResNetControl


class Classifier:
    def __init__(self):
        self.modelhub = ModelHub.from_github_repo()
        self.modelhub.parse_objects()

        self.response: List[bool | None] = []
        self.prompt: str = ""
        self.label: str = ""
        self.model_name: str = ""

    def _parse_label(self, prompt: str):
        self.prompt = prompt
        lang = "zh" if re.compile("[\u4e00-\u9fa5]+").search(prompt) else "en"
        _label = label_cleaning(prompt)
        _label = split_prompt_message(_label, lang=lang)
        self.label = _label

    def rank_models(
        self, nested_models: List[str], example_paths: List[Path | bytes]
    ) -> ResNetControl | None:
        # {{< Rank ResNet Models >}}
        rank_ladder = []
        for example_path in example_paths:
            if isinstance(example_path, bytes):
                img_stream = example_path
            elif isinstance(example_path, Path):
                img_stream = example_path.read_bytes()
            else:
                continue

            for model_name in nested_models:
                net = self.modelhub.match_net(focus_name=model_name)
                control = ResNetControl.from_pluggable_model(net)
                result_, proba = control.execute(img_stream, proba=True)
                if result_:
                    rank_ladder.append([control, model_name, proba])
                    if proba[0] > 0.87:
                        break

        # {{< Catch-all Rules >}}
        if rank_ladder:
            alts = sorted(rank_ladder, key=lambda x: x[-1][0], reverse=True)
            best_model, model_name = alts[0][0], alts[0][1]
            self.model_name = model_name
            return best_model

    def inference(self, images: List[Path | bytes], control: ResNetControl):
        for image in images:
            try:
                if isinstance(image, Path):
                    if not image.exists():
                        self.response.append(None)
                        continue
                    image = image.read_bytes()
                if isinstance(image, bytes):
                    result = control.execute(image)
                    self.response.append(result)
                else:
                    self.response.append(None)
            except Exception as err:
                logger.debug(str(err), prompt=self.prompt)
                self.response.append(None)

    def execute(
        self,
        prompt: str,
        images: List[Path | bytes],
        example_paths: List[Path | bytes] | None = None,
    ) -> List[bool | None]:
        self.response = []

        self._parse_label(prompt)

        # Match: model ranker
        if nested_models := self.modelhub.nested_categories.get(self.label, []):
            if model := self.rank_models(nested_models, example_paths):
                self.inference(images, model)
        # Match: common binary classification
        elif focus_label := self.modelhub.label_alias.get(self.label):
            self.model_name = (
                focus_label if focus_label.endswith(".onnx") else f"{focus_label}.onnx"
            )
            net = self.modelhub.match_net(self.model_name)
            control = ResNetControl.from_pluggable_model(net)
            self.inference(images, control)
        # Match: Unknown cases
        else:
            logger.debug("Types of challenges not yet scheduled", label=self.label, prompt=prompt)

        return self.response


class LocalBinaryClassifier:
    def __init__(self, model_path: Path):
        if not isinstance(model_path, Path) or not model_path.exists():
            raise FileNotFoundError
        self.model_path = model_path
        net = cv2.dnn.readNetFromONNX(str(model_path))
        self.model = ResNetControl.from_pluggable_model(net)

    def parse_once(self, image: bytes) -> bool | None:
        with suppress(Exception):
            return self.model.execute(image)



# -*- coding: utf-8 -*-
# Time       : 2023/9/29 14:19
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from typing import Tuple, List

import cv2
import numpy as np
from skimage.metrics import structural_similarity as compare_ssim
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import rbf_kernel


def limited_radius(img) -> int:
    # x1, y1, x2, y2 = 90, 200, 413, 523
    r = 48 if 512 in img.shape else 32
    return r


def annotate_objects(image_path: str):
    img = cv2.imread(image_path)
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    circles = cv2.HoughCircles(
        gray, cv2.HOUGH_GRADIENT, 1, 66, param1=100, param2=40, minRadius=5, maxRadius=65
    )

    if np.all(circles):
        circles = np.uint16(np.around(circles))
        ix = []
        for i in circles[0, :]:
            lr = limited_radius(img)
            if i[2] > lr + 2:
                continue
            i[2] = lr
            ix.append(i)
        return img, ix

    return img, None


def find_unique_object(img: np.ndarray, circles: List[List[int]]) -> Tuple[int, int, int]:
    def spectral_clustering():
        num_clusters = 2

        S = rbf_kernel(feature_vectors)
        D = np.diag(np.sum(S, axis=1))
        L = D - S

        eigvals, eigvecs = np.linalg.eigh(L)
        eigvecs = eigvecs[:, :num_clusters]

        kmeans = KMeans(n_clusters=num_clusters, n_init=10)
        kmeans.fit(eigvecs)

        return kmeans.labels_

    if len(circles) == 1:
        x, y, r = circles[0]
        return x, y, r

    feature_vectors = []
    for x, y, r in circles:
        roi = img[y - r : y + r, x - r : x + r]
        hist = cv2.calcHist([roi], [0, 1, 2], None, [16, 16, 16], [0, 256, 0, 256, 0, 256])
        hist = cv2.normalize(hist, hist).flatten()
        feature_vectors.append(hist)

    labels = spectral_clustering()

    unique, counts = np.unique(labels, return_counts=True)
    car_1_label = unique[np.argmin(counts)]

    for i, label in enumerate(labels):
        if label == car_1_label:
            x, y, r = circles[i][0:3]
            return x, y, r


def find_unique_object_v2(img: np.ndarray, circles: List[List[int]]) -> Tuple[int, int, int]:
    mask_images = []

    for circle in circles:
        x, y, r = circle
        mask = np.zeros(img.shape[:2], dtype=np.uint8)
        mask = cv2.circle(mask, (x, y), r, (255, 255, 0), -1)
        mask_img = cv2.bitwise_and(img, img, mask=mask)
        mask_img = mask_img[y - r : y + r, x - r : x + r]
        mask_images.append(mask_img)

    max_size = max([mask_img.shape[0] for mask_img in mask_images])
    mask_images = [cv2.resize(mask_img, (max_size, max_size)) for mask_img in mask_images]

    similarity = []

    for i, mask_img in enumerate(mask_images):
        sig_sim = []
        for j, mask_img_ in enumerate(mask_images):
            if i == j:
                sig_sim.append(0)
                continue
            score, _ = compare_ssim(mask_img, mask_img_, win_size=3, full=True)
            sig_sim.append(score)
        similarity.append(np.array(sig_sim))

    similarity = np.array(similarity)
    sum_similarity = np.sum(similarity, axis=0)
    unique_index = np.argmin(sum_similarity)
    unique_circle = circles[unique_index]

    return unique_circle[0], unique_circle[1], unique_circle[2]



# -*- coding: utf-8 -*-
# Time       : 2023/10/9 16:17
# Author     : QIN2DIM
# GitHub     : https://github.com/QIN2DIM
# Description:
from .appears_only_once import annotate_objects, find_unique_object, find_unique_object_v2
from .largest_animal import find_similar_objects

__all__ = [
    "annotate_objects",
    "find_unique_object",
    "find_similar_objects",
    "find_unique_object_v2",
]
